class BaseContext(object):
    def __init__(self, evaluator, parent_context=None):
        self.evaluator = evaluator
        self.parent_context = parent_context

    def get_root_context(self):
        context = self
        while True:
            if context.parent_context is None:
                return context
            context = context.parent_context


class BaseContextSet(object):
    def __init__(self, *args):
        self._set = set(args)

    @classmethod
    def from_iterable(cls, iterable):
        return cls.from_set(set(iterable))

    @classmethod
    def from_set(cls, set_):
        self = cls()
        self._set = set_
        return self

    @classmethod
    def from_sets(cls, sets):
        """
        Used to work with an iterable of set.
        """
        aggregated = set()
        sets = list(sets)
        for set_ in sets:
            if isinstance(set_, BaseContextSet):
                aggregated |= set_._set
            else:
                aggregated |= set_
        return cls.from_set(aggregated)

    def __or__(self, other):
        return type(self).from_set(self._set | other._set)

    def __iter__(self):
        for element in self._set:
            yield element

    def __bool__(self):
        return bool(self._set)

    def __len__(self):
        return len(self._set)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join(str(s) for s in self._set))

    def filter(self, filter_func):
        return type(self).from_iterable(filter(filter_func, self._set))

    def __getattr__(self, name):
        def mapper(*args, **kwargs):
            return type(self).from_sets(
                getattr(context, name)(*args, **kwargs)
                for context in self._set
            )
        return mapper



from jedi.evaluate.base_context import ContextSet, NO_CONTEXTS

class AbstractLazyContext(object):
    def __init__(self, data):
        self.data = data

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.data)

    def infer(self):
        raise NotImplementedError


class LazyKnownContext(AbstractLazyContext):
    """data is a context."""
    def infer(self):
        return ContextSet(self.data)


class LazyKnownContexts(AbstractLazyContext):
    """data is a ContextSet."""
    def infer(self):
        return self.data


class LazyUnknownContext(AbstractLazyContext):
    def __init__(self):
        super(LazyUnknownContext, self).__init__(None)

    def infer(self):
        return NO_CONTEXTS


class LazyTreeContext(AbstractLazyContext):
    def __init__(self, context, node):
        super(LazyTreeContext, self).__init__(node)
        self._context = context
        # We need to save the predefined names. It's an unfortunate side effect
        # that needs to be tracked otherwise results will be wrong.
        self._predefined_names = dict(context.predefined_names)

    def infer(self):
        old, self._context.predefined_names = \
            self._context.predefined_names, self._predefined_names
        try:
            return self._context.eval_node(self.data)
        finally:
            self._context.predefined_names = old


def get_merged_lazy_context(lazy_contexts):
    if len(lazy_contexts) > 1:
        return MergedLazyContexts(lazy_contexts)
    else:
        return lazy_contexts[0]


class MergedLazyContexts(AbstractLazyContext):
    """data is a list of lazy contexts."""
    def infer(self):
        return ContextSet.from_sets(l.infer() for l in self.data)
